--- old/usr/src/uts/common/fs/zfs/zfs_ioctl.c	2012-07-30 16:07:12.759607272 +0400
+++ new/usr/src/uts/common/fs/zfs/zfs_ioctl.c	2012-07-30 16:07:12.219607259 +0400
@@ -3227,21 +3227,32 @@
 	int error;
 	zfsvfs_t *zfsvfs;
 	char *clone_name;
+	boolean_t umounted = B_FALSE;
+
+	if (getzfsvfs(zc->zc_name, &zfsvfs) == 0) {
+		error = zfs_suspend_fs(zfsvfs);
+		if (error) {
+			VFS_RELE(zfsvfs->z_vfs);
+			return (error);
+		}
+
+		umounted = B_TRUE;
+	}
 
 	error = dsl_dataset_hold(zc->zc_name, FTAG, &ds);
 	if (error)
-		return (error);
+		goto out_vfs;
 
 	/* must not be a snapshot */
 	if (dsl_dataset_is_snapshot(ds)) {
-		dsl_dataset_rele(ds, FTAG);
-		return (EINVAL);
+		error = EINVAL;
+		goto out_ds;
 	}
 
 	/* must have a most recent snapshot */
 	if (ds->ds_phys->ds_prev_snap_txg < TXG_INITIAL) {
-		dsl_dataset_rele(ds, FTAG);
-		return (EINVAL);
+		error = EINVAL;
+		goto out_ds;
 	}
 
 	/*
@@ -3259,23 +3270,15 @@
 	/*
 	 * Do clone swap.
 	 */
-	if (getzfsvfs(zc->zc_name, &zfsvfs) == 0) {
-		error = zfs_suspend_fs(zfsvfs);
-		if (error == 0) {
-			int resume_err;
-
-			if (dsl_dataset_tryown(ds, B_FALSE, FTAG)) {
-				error = dsl_dataset_clone_swap(clone, ds,
-				    B_TRUE);
-				dsl_dataset_disown(ds, FTAG);
-				ds = NULL;
-			} else {
-				error = EBUSY;
-			}
-			resume_err = zfs_resume_fs(zfsvfs, zc->zc_name);
-			error = error ? error : resume_err;
+	if (umounted) {
+		if (dsl_dataset_tryown(ds, B_FALSE, FTAG)) {
+			error = dsl_dataset_clone_swap(clone, ds,
+			    B_TRUE);
+			dsl_dataset_disown(ds, FTAG);
+			ds = NULL;
+		} else {
+			error = EBUSY;
 		}
-		VFS_RELE(zfsvfs->z_vfs);
 	} else {
 		if (dsl_dataset_tryown(ds, B_FALSE, FTAG)) {
 			error = dsl_dataset_clone_swap(clone, ds, B_TRUE);
@@ -3293,8 +3296,18 @@
 
 out:
 	strfree(clone_name);
+out_ds:
 	if (ds)
 		dsl_dataset_rele(ds, FTAG);
+out_vfs:
+	if (umounted) {
+		int resume_err;
+
+		resume_err = zfs_resume_fs(zfsvfs, zc->zc_name);
+		error = error ? error : resume_err;
+		VFS_RELE(zfsvfs->z_vfs);
+	}
+
 	return (error);
 }
 
@@ -3701,7 +3714,12 @@
 	if (props) {
 		nvlist_t *errlist;
 
-		if (dmu_objset_from_ds(drc.drc_logical_ds, &os) == 0) {
+		if (drc.drc_newfs)
+			error = dmu_objset_from_ds(drc.drc_real_ds, &os);
+		else
+			error = dmu_objset_hold(tofs, FTAG, &os);
+
+		if (error == 0) {
 			if (drc.drc_newfs) {
 				if (spa_version(os->os_spa) >=
 				    SPA_VERSION_RECVD_PROPS)
@@ -3714,6 +3732,9 @@
 				zc->zc_obj |= ZPROP_ERR_NOCLEAR;
 			}
 			dsl_prop_set_hasrecvd(os);
+
+			if (!drc.drc_newfs)
+				dmu_objset_rele(os, FTAG);
 		} else if (!drc.drc_newfs) {
 			zc->zc_obj |= ZPROP_ERR_NOCLEAR;
 		}
